@@ -187,11 +187,16 @@ def _map_status(hvac_state) -> Tuple[str, str]:
     return online, update
 
 
-def map_system_info(facilities, gateway, hvac) -> SystemInfo:
+def map_system_info(facilities, gateway, hvac, serial) -> SystemInfo:
     """Map *system info*."""
-    serial = map_serial_number(facilities)
+    if serial is None:
+        serial = map_serial_number(facilities)
+
+    facilities_list = facilities.get(
+        "body", dict()).get("facilitiesList", list())
+    facility = [facility for facility in
+                facilities_list if facility.get('serialNumber') == serial][0]
 
-    facility = facilities.get("body", dict()).get("facilitiesList", list())[0]
     name = facility.get("name", None)
     mac_ethernet = facility.get("networkInformation", dict()) \
         .get("macAddressEthernet")
@@ -230,7 +235,8 @@ def map_zone(raw_zone) -> Optional[Zone]:
         configuration = raw_zone.get("configuration", dict())
         name = configuration.get("name", "").strip()
         temperature = configuration.get("inside_temperature")
-        active_function = ActiveFunction[configuration.get("active_function")]
+        active_function = ActiveFunction[
+            configuration.get("active_function", ActiveFunction.STANDBY.name)]
         quick_veto = _map_quick_veto_zone(configuration.get("quick_veto"))
         rbr = raw_zone.get("currently_controlled_by", dict())\
             .get("name", "") == "RBR"
